package de.uniulm.mathematik.typo.fonts.afm;

import java.lang.*;

/**
 * An instance of this class represents the metrics of an individual
 * glyph of a font.
 * The metrics includes the character code. This code value can be -1 in
 * case of glyphs that are not included in the encoding vector.
 * A name is usually specified but this is not required. Likewise,
 * the width vectors and the bounding box are optional. Ligatures
 * are parsed but currently not yet stored.
 * See also: Adobe Font Metrics File Specification, chapter 8.
 */
public class CharacterMetrics {
   private int code;
   private String line;
   private PropertyTable properties; 
   private String name;
   private Lex lex;

   /**
    * Inner class for exceptions raised by the CharacterMetrics class.
    */
   public static class ParserException extends Exception {
      /**
       * The corresponding input line of the AFM source.
       */
      public final String line;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * source line and a printable message within a throw statement.
       * @param line source input line with the character metrics.
       * @param message printable error message.
       */
      public ParserException (String line, String message) {
	 this.line = line;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the source line and the error message.
       * @return the string with the error message.
       */
      public String toString() {
	 return message + ": " + line;
      }
   } // class ParserException

   /**
    * Private inner class for the lexical analysis of character metrics.
    */
   private class Lex {
      private String line;
      private int index;
      private char ch;
      private boolean eof;
      public final static int EOF = 0;
      public final static int ERROR = 1;
      public final static int SEMICOLON = 2;
      public final static int IDENTIFIER = 3;
      public final static int INTEGER = 4;
      /**
       * Last symbol scanned by getsy().
       */
      public int sy;
      /**
       * Identifier string if sy equals IDENTIFER.
       */
      public String ident;
      /**
       * Integer value if sy equals INTEGER.
       */
      public int intval;

      /**
       * Advance to the next character of the input line. This
       * method sets the private variable ch.
       */
      private void nextch() {
	 if (!eof) {
	    if (index == line.length()) {
	       eof = true;
	    } else {
	       ch = line.charAt(index);
	       ++index;
	    }
	 }
      }

      /**
       * Scan the next symbol from the input line. Possible symbol
       * values are EOF, ERROR, SEMICOLON, IDENTIFER, and INTEGER.
       * The symbol value is stored in the public variable sy.
       */
      public void getsy() {
	 while (!eof && Character.isWhitespace(ch)) {
	    nextch();
	 }
	 if (eof) {
	    sy = EOF;
	 } else if (ch == ';') {
	    nextch();
	    sy = SEMICOLON;
	 } else if (Character.isLetter(ch) || ch == '.') {
	    ident = (new Character(ch)).toString();
	    nextch();
	    while (!eof && !Character.isWhitespace(ch)) {
	       ident += ch;
	       nextch();
	    }
	    sy = IDENTIFIER; 
	 } else if (ch == '-' || Character.isDigit(ch)) {
	    String value = (new Character(ch)).toString();
	    nextch();
	    while (!eof && Character.isDigit(ch)) {
	       value += ch;
	       nextch();
	    }
	    try {
	       intval = java.lang.Integer.parseInt(value, 10);
	       sy = INTEGER;
	    } catch (NumberFormatException e) {
	       sy = ERROR;
	    }
	 } else if (ch == '<') {
	    String value = "";
	    nextch();
	    while (!eof && ch != '>') {
	       value += ch;
	       nextch();
	    }
	    try {
	       intval = java.lang.Integer.parseInt(value, 16);
	       sy = INTEGER;
	    } catch (NumberFormatException e) {
	       sy = ERROR;
	    }
	    nextch();
	 } else {
	    sy = ERROR;
	 }
      }

      /**
       * Construct lexical analyser for the given source line.
       * @param line source line.
       */
      public Lex(String line) {
	 this.line = line;
	 index = 0;
	 eof = false;
	 nextch();
      }
   }

   /**
    * Raise an exception if lex.sy does not equal
    * the given symbol value.
    * @param symbol expected symbol.
    * @param what string explaining what is expected. This is
    *             used for error messages.
    */
   private void expect(int symbol, String what) throws ParserException {
      if (lex.sy != symbol) {
	 throw new ParserException(line, what + " expected");
      }
   }

   /**
    * Expect and return an integer from the lexical analysis.
    * The next input symbol is read, if successful.
    * @param what string explaining what is epxected. This is
    *             used for error messages.
    * @return integer value read from the input.
    */
   private int fetchint(String what) throws ParserException {
      expect(Lex.INTEGER, "integer");
      int value = lex.intval;
      lex.getsy();
      return value;
   }

   /**
    * Expect and return an identifier from the lexical analysis.
    * The next input symbol is read, if successful.
    * @param what string explaining what is epxected. This is
    *             used for error messages.
    * @return string of identifier as read from the input.
    */
   private String fetchident(String what) throws ParserException {
      expect(Lex.IDENTIFIER, what);
      String ident = lex.ident;
      lex.getsy();
      return ident;
   }

   /**
    * Add property/value pair to the list of properties belonging to
    * this glyph.
    * @param property The name of the property. This is usually the
    *                 keyword as found in the AFM source file.
    * @param object   The associated value, either an Integer object
    *                 or a string.
    */
   private void addProperty(String property, Object object)
	 throws ParserException {
      try {
	 properties.add(property, object);
      } catch (PropertyTable.PropertyTableException e) {
	 throw new ParserException(line, e.toString());
      }
   }

   /**
    * Add integer values property/value pair to the list of properties
    * belonging to this glyph.
    * @param property The name of the property. This is usually the
    *                 keyword as found in the AFM source file.
    * @param value    The associated integer value.
    */
   private void addIntegerProperty(String property, int value)
	 throws ParserException {
      addProperty(property, new Integer(value));
   }

   /**
    * Parse an individual property entry within the input source line.
    */
   private void parseProperty() throws ParserException {
      String property = fetchident("property");
      if (property.equals("C") || property.equals("CX")) {
	 code = fetchint(property); addIntegerProperty("C", code);
      } else if (property.equals("W")) {
	 int wx = fetchint("WX"); int wy = fetchint("WY");
	 addIntegerProperty("WX", wx); addIntegerProperty("WY", wy);
      } else if (property.equals("W0")) {
	 int wx = fetchint("W0X"); int wy = fetchint("W0Y");
	 addIntegerProperty("W0X", wx); addIntegerProperty("W0Y", wy);
      } else if (property.equals("W1")) {
	 int wx = fetchint("W1X"); int wy = fetchint("W1Y");
	 addIntegerProperty("W1X", wx); addIntegerProperty("W1Y", wy);
      } else if (property.equals("WX")) {
	 int wx = fetchint("WX");
	 addIntegerProperty("WX", wx); addIntegerProperty("WY", 0);
      } else if (property.equals("WY")) {
	 int wy = fetchint("WY");
	 addIntegerProperty("WY", wy); addIntegerProperty("WX", 0);
      } else if (property.equals("VV")) {
	 int vvx = fetchint("VVX"); int vvy = fetchint("VVY");
	 addIntegerProperty("VVX", vvx); addIntegerProperty("VVY", vvy);
      } else if (property.equals("N")) {
	 name = fetchident("name");
	 if (!name.equals(".notdef")) {
	    addProperty("N", name);
	 }
      } else if (property.equals("B")) {
	 int llx = fetchint("B llx"); int lly = fetchint("B lly");
	 int urx = fetchint("B urx"); int ury = fetchint("B ury");
	 addIntegerProperty("B.llx", llx);
	 addIntegerProperty("B.lly", lly);
	 addIntegerProperty("B.urx", urx);
	 addIntegerProperty("B.ury", ury);
      } else if (property.equals("L")) {
	 String successor = fetchident("successor");
	 String ligature = fetchident("ligature");
      }
   }

   /**
    * Parse the entire source line.
    */
   private void parse() throws ParserException {
      parseProperty();
      while (lex.sy == Lex.SEMICOLON) {
	 lex.getsy();
	 if (lex.sy != Lex.SEMICOLON && lex.sy != Lex.EOF) {
	    parseProperty();
	 }
      }
      if (lex.sy != Lex.EOF) {
	 throw new ParserException(line, "syntax error");
      }
   }

   /**
    * Return the internal data structures as printable string.
    */
   public String toString() {
      if (name != null) {
	 return "character metrics of " + name + ": " + properties.toString();
      } else {
	 return "character metrics " + properties.toString();
      }
   }

   /**
    * Return the name of this glyph. Please note that the name may
    * be null.
    */
   public String getName() {
      return name;
   }

   /**
    * Return the code value. This value is within the range 0..255,
    * if this glyph is included by the encoding vector. Otherwise,
    * a value of -1 is returned.
    */
   public int getCode() {
      return code;
   }

   /**
    * Return true if the given property is defined for this glyph.
    * @param property name of property, e.g. "WX".
    */
   public boolean defined(String property) {
      return properties.defined(property);
   }

   /**
    * Return the integer value for this property of this glyph.
    * @param property name of property. Possible names are
    *                 C, WX, WY, W0X, W0Y, W1X, W1Y, VVX, VVY,
    *                 B.llx, B.lly, B.urx, and B.ury.
    */
   public int get(String property) throws PropertyTable.PropertyTableException {
      Integer value = (Integer) properties.get(property);
      return value.intValue();
   }

   /**
    * Return the width in x direction. 0 is returned, if no width was
    * specified.
    */
   public int getWX() {
      try {
	 return get("WX");
      } catch(PropertyTable.PropertyTableException e) {
	 return 0;
      }
   }

   /**
    * Return the width in y direction. 0 is returned, if no width was
    * specified.
    */
   public int getWY() {
      try {
	 return get("WY");
      } catch(PropertyTable.PropertyTableException e) {
	 return 0;
      }
   }

   /**
    * Return the bounding box as an array of four integer values,
    * representing the values llx, lly, urx, and ury. Four zeroes
    * are returned if no bounding box was specified.
    */
   public int[] getBoundingBox() {
      int[] bbox = new int[4];
      try {
	 bbox[0] = get("B.llx");
	 bbox[1] = get("B.lly");
	 bbox[2] = get("B.urx");
	 bbox[3] = get("B.ury");
      } catch(PropertyTable.PropertyTableException e) {
	 for (int i = 0; i < 4; ++i) {
	    bbox[i] = 0;
	 }
      }
      return bbox;
   }

   /**
    * Construct a character metrics object from a AFM source line.
    * @param line The AFM source line.
    */
   public CharacterMetrics(String line) throws ParserException {
      this.line = line;
      name = null;
      code = -1;
      properties = new PropertyTable();
      lex = new Lex(line);
      lex.getsy();
      parse();
   }
} // class CharacterMetrics
