package de.uniulm.mathematik.typo.fonts.afm;

/**
 * Instances of this class represent parser objects for AFM source
 * files that are expected to conform to following grammar:
 * <P>
 * <UL>
 * <LI>FontMetrics --&gt; START VALUE ListOfProperties END</LI>
 * <LI>ListOfProperties --&gt;</LI>
 * <LI>ListOfProperties --&gt; Property ListOfProperties</LI>
 * <LI>Property --&gt; SimpleProperty</LI>
 * <LI>Property --&gt; Block</LI>
 * <LI>Property --&gt; KernData</LI>
 * <LI>SimpleProperty --&gt; PROPERTY OptionalValue</LI>
 * <LI>KernData --&gt; KERN_DATA_START ListOfBlocks KERN_DATA_END</LI>
 * <LI>ListOfBlocks --&gt; Block</LI>
 * <LI>ListOfBlocks --&gt; Block ListOfBlocks</LI>
 * <LI>Block --&gt; BLOCK_START OptionalValue ListOfBlockLines BLOCK_END</LI>
 * <LI>OptionalValue --&gt;</LI>
 * <LI>OptionalValue --&gt; VALUE</LI>
 * <LI>ListOfBlockLines --&gt;</LI>
 * <LI>ListOfBlockLines --&gt; BLOCK_LINE ListOfBlockLines</LI>
 * </UL>
 * <P>
 * Individual entries of character metrics are handled by
 * the CharacterMetrics class. Kerning pairs are parsed by the
 * KerningPair class.
 * <P>
 * See Adobe Font Metrics File Specification, chapter 3.
 */
public class Parser {
   private final Lex lex;
   private Lex.Token token;
   private PropertyTable properties;
   private CharacterMetricsTable cms;
   private KerningPairTable kpairs;
   private boolean nextToken;
   private String version;
   private int nestlevel; // nestlevel of blocks

   /**
    * Exception class for syntax errors detected by the parser.
    * The two instance variables token and message are final
    * and public.
    *
    * A more detailed report including the exact error position
    * is not supported yet.
    */
   public static class ParserException extends Exception {
      /**
       * The current token when the syntax error was detected.
       */
      public final Lex.Token token;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * token and a printable message within a throw statement.
       * @param token current token.
       * @param message printable error message.
       */
      public ParserException (Lex.Token token, String message) {
	 this.token = token;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the token and the error message.
       * @return the string with the error message.
       */
      public String toString() {
	 return "at " + token + ": " + message;
      }
   } // class ParserException

   /**
    * Construct a parser object.
    * @param lex lexical feed of tokens.
    */
   public Parser(Lex lex) throws ParserException {
      this.lex = lex;
      properties = new PropertyTable();
      cms = new CharacterMetricsTable();
      kpairs = new KerningPairTable();
      nextToken = true;
      parse();
      kpairs.resolve(cms);
   }

   /**
    * Return table of properties.
    */
   public PropertyTable getProperties() {
      return properties;
   }

   /**
    * Return table of character metrics. The table is empty
    * if no character metrics entries were included in the source.
    */
   public CharacterMetricsTable getCharacterMetrics() {
      return cms;
   }

   /**
    * Return table of kerning pairs. The table is empty
    * if no kerning pairs were specified.
    */
   public KerningPairTable getKerningPairs() {
      return kpairs;
   }

   /**
    * Return version number that was given along with the StartFontMetrics
    * keyword.
    */
   public String getVersion() {
      return version;
   }

   /**
    * Return internal data structures as printable string.
    */
   public String toString() {
      return "Version: " + version + "\n" +
	 "Properties: " + properties.toString() + "\n" +
	 "CharacterMetrics: " + cms.toString() + "\n" +
	 "KerningPairs: " + kpairs.toString() + "\n";
   }

   /**
    * Read the next token on the next invocation of
    * getToken().
    */
   private void nextToken() {
      if (nextToken) {
	 token = lex.getToken();
      } else {
	 nextToken = true;
      }
   }

   /**
    * Return the current token.
    */
   private Lex.Token getToken() {
      if (nextToken) {
	 token = lex.getToken();
	 nextToken = false;
      }
      return token;
   }

   private void expect(int symbol, String what) throws ParserException {
      if (getToken().symbol != symbol) {
	 throw new ParserException(getToken(), what + " expected");
      }
   }

   private void skip(int symbol, String what) throws ParserException {
      expect(symbol, what);
      nextToken();
   }

   private void addProperty(String property, Object object)
	 throws ParserException {
      try {
	 properties.add(property, object);
      } catch (PropertyTable.PropertyTableException e) {
	 throw new ParserException(getToken(), e.toString());
      }
   }

   private void parseSimpleProperty() throws ParserException {
      expect(Lex.Token.PROPERTY, "property");
      String property = (String) getToken().object;
      nextToken();
      if (getToken().symbol == Lex.Token.VALUE) {
	 addProperty(property, getToken().object);
	 nextToken();
      } else {
	 addProperty(property, null);
      }
   }

   private void parseCharacterMetrics() throws ParserException {
      while (getToken().symbol == Lex.Token.BLOCK_LINE) {
	 String line = (String) getToken().object;
	 try {
	    CharacterMetrics metrics = new CharacterMetrics(line);
	    cms.add(metrics);
	 } catch (Exception e) {
	    throw new ParserException(getToken(), e.toString());
	 }
	 nextToken();
      }
   }

   private void parseKerningPairs() throws ParserException {
      while (getToken().symbol == Lex.Token.BLOCK_LINE) {
	 String line = (String) getToken().object;
	 try {
	    KerningPair pair = new KerningPair(line);
	    kpairs.add(pair);
	 } catch (Exception e) {
	    throw new ParserException(getToken(), e.toString());
	 }
	 nextToken();
      }
   }

   private void parseBlock() throws ParserException {
      ++nestlevel;
      expect(Lex.Token.BLOCK_START, "start of a block");
      String block = (String) getToken().object;
      nextToken();
      expect(Lex.Token.VALUE, "value of " + block);
      nextToken();
      if (nestlevel == 1 && block.equals("CharMetrics")) {
	 parseCharacterMetrics();
      } else if (nestlevel == 2 && block.equals("KernPairs")) {
	 parseKerningPairs();
      } else {
	 // skip this block
	 while (getToken().symbol == Lex.Token.BLOCK_LINE) {
	    nextToken();
	 }
      }
      expect(Lex.Token.BLOCK_END, "End" + block);
      nextToken();
      --nestlevel;
   }

   private void parseListOfBlocks() throws ParserException {
      while (getToken().symbol == Lex.Token.BLOCK_START) {
	 parseBlock();
      }
   }

   // KernData --> KERN_DATA_START ListOfBlocks KERN_DATA_END
   private void parseKernData() throws ParserException {
      ++nestlevel;
      skip(Lex.Token.KERN_DATA_START, "StartKernData");
      parseListOfBlocks();
      skip(Lex.Token.KERN_DATA_END, "EndKernData");
      --nestlevel;
   }

   private void parseListOfProperties() throws ParserException {
      for(;;) {
	 switch(getToken().symbol) {
	    case Lex.Token.PROPERTY:
	       parseSimpleProperty();
	       break;
	    case Lex.Token.BLOCK_START:
	       parseBlock();
	       break;
	    case Lex.Token.KERN_DATA_START:
	       parseKernData();
	       break;
	    default:
	       return;
	 }
      }
   }

   private void parse() throws ParserException {
      skip(Lex.Token.START, "StartFontMetrics");
      expect(Lex.Token.VALUE, "version number");
      version = (String) getToken().object;
      nestlevel = 0;
      nextToken();
      parseListOfProperties();
      skip(Lex.Token.END, "EndFontMetrics");
      expect(Lex.Token.EOF, "EOF");
   }
} // class Parser
