package de.uniulm.mathematik.typo.hyphen;

import java.util.*;

/**
 * Hyphenation entries represent a possibly shared
 * sequence of hyphenation points. They are hashable
 * such that common suffixes of hyphenation point sequences
 * can be found using a hash table.
 */
public class HyphenationEntry
      implements HyphenationRule, HyphenationPoint {

   protected int pos;
   protected int value;
   protected HyphenationEntry next;

   /**
    * This iterators allows to iterate through the hyphenation point
    * beginning at this hyphenation entry.
    */
   public class HyphenationEntryIterator
	 implements Iterator<HyphenationPoint> {

      private HyphenationEntry entry;

      /**
       * Create iterator beginning at e.
       */
      public HyphenationEntryIterator(HyphenationEntry e) {
	 entry = e;
      }

      /**
       * Return true if there is at least one value left
       * which can be returned by the iterator.
       */
      public boolean hasNext() {
	 return entry != null;
      }

      /**
       * Not supported, raises UnsupportedOperationException.
       */
      public void remove() {
	 throw new UnsupportedOperationException();
      }

      /**
       * Return next hyphenation entry.
       * This method raises NoSuchElementException if hasNext()
       * would have returned false.
       */
      public HyphenationEntry next() {
	 if (entry == null) {
	    throw new NoSuchElementException();
	 }
	 HyphenationEntry e = entry;
	 entry = entry.next;
	 return e;
      }
   }

   /**
    * Return the first hyphenation position.
    */
   public int getPosition() {
      return pos;
   }

   /**
    * Return the first hyphenation value. Odd values represent
    * positions where a hyphenation is desired, even values indicate
    * a suppression of a hyphenation.
    */
   public int getValue() {
      return value;
   }

   /**
    * Return iterator that allows to examine all hyphenation points
    * of this hyphenation entry.
    */
   public Iterator<HyphenationPoint> iterator() {
      return new HyphenationEntryIterator(this);
   }

   /**
    * Returns true iff object is an HyphenationEntry with an
    * equivalent sequence of hyphenation positions and values.
    */
   public boolean equals(Object object) {
      if (object instanceof HyphenationEntry) {
	 HyphenationEntry entry1 = this;
	 HyphenationEntry entry2 = (HyphenationEntry) object;
	 if (entry1 == entry2) return true;
	 for(;;) {
	    if (entry1.pos != entry2.pos || entry1.value != entry2.value) {
	       return false;
	    }
	    if ((entry1.next == null) != (entry2.next == null)) {
	       return false;
	    }
	    if (entry1.next == null) return true;
	    entry1 = entry1.next; entry2 = entry2.next;
	 }
      } else {
	 return false;
      }
   }

   /**
    * Returns a hash code that depends on the sequence
    * of hyphenation positions and values.
    */
   public int hashCode() {
      int hashval = 0;
      for (HyphenationEntry entry = this; entry != null; entry = entry.next) {
	 hashval = Integer.rotateLeft(hashval, 3)
	    ^ (entry.pos << 4) ^ entry.value;
      }
      return hashval;
   }

   /**
    * Create a new hyphenation entry with the first hyphenation point
    * defined by p and v. The remaining hyphenation points are given
    * by n, if non-null.
    */
   protected HyphenationEntry(int p, int v, HyphenationEntry n) {
      assert n == null || p < n.pos;
      pos = p; value = v; next = n;
   }

   /**
    * Returns a string that contains the comma-separated
    * sequence of hyphenation positions and their values,
    * enclosed in braces. This is useful for debugging purposes.
    */
   public String toString() {
      StringBuffer result = new StringBuffer();
      result.append("{");
      HyphenationEntry entry = this;
      boolean first = true;
      while (entry != null) {
	 if (first) {
	    first = false;
	 } else {
	    result.append(", ");
	 }
	 result.append(entry.pos + ":" + entry.value);
	 entry = entry.next;
      }
      result.append("}");
      return result.toString();
   }
}
