package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * Class with the static method fit() that fits a horizontal sequence
 * to a given length.
 */
public class HorizontalFitter {

   private interface Adapter {
      int getMaximalAmount();
      void adapt(int amount);
   }

   private static class Stretcher implements Adapter {
      private Item item;

      public Stretcher(Item item) {
	 this.item = item;
      }

      public int getMaximalAmount() {
	 return item.getStretchability();
      }

      public void adapt(int amount) {
	 item.stretch(amount);
      }
   }

   private static class Shrinker implements Adapter {
      private Item item;

      public Shrinker(Item item) {
	 this.item = item;
      }

      public int getMaximalAmount() {
	 return item.getShrinkability();
      }

      public void adapt(int amount) {
	 item.shrink(amount);
      }
   }

   private static void adaptUniformely(Iterable<Adapter> items,
	 int count, int amount) {
      if (count > 0) {
	 int share = amount / count;
	 int remainder = amount % count;
	 for (Adapter adapter: items) {
	    int change = share;
	    if (remainder > 0) {
	       change++; --remainder;
	    }
	    adapter.adapt(change);
	 }
      }
   }

   private static void stretchinf(HorizontalSequence hseq, int amount) {
      LinkedList<Adapter> items = new LinkedList<Adapter>();
      for (Item item: hseq.getGlueItems()) {
	 if (item.getStretchability() == Item.INFINITY) {
	    items.addLast(new Stretcher(item));
	 }
      }
      assert(items.size() > 0);
      adaptUniformely(items, items.size(), amount);
   }

   private static void adaptWeighted(Iterable<Adapter> items,
	 int count, int amount, int maximalAmount) {
      assert(count > 0);
      double ratio = (double) amount / maximalAmount;
      int total = 0;
      for (Adapter adapter: items) {
	 int change = (int) (adapter.getMaximalAmount() * ratio);
	 adapter.adapt(change);
	 total += change;
      }
      if (total < amount) {
	 for (Adapter adapter: items) {
	    adapter.adapt(1);
	    if (--total == 0) {
	       break;
	    }
	 }
      }
   }

   /**
    * Adapt the given horizontal sequence such that it will be shrank
    * or stretched to the given horizontal width. This method will
    * not shrink the sequence beyond the specified shrinkability.
    * However, the specified stretchability is ignored.
    * @param hseq the to be adapted horizontal sequence.
    * @param width the intended horizontal width.
    */
   public static void fit(HorizontalSequence hseq, int width) {
      Width w = hseq.getWidth();
      if (w.getWidth() == width) return;
      if (w.getWidth() < width && w.getStretchability() == Item.INFINITY) {
	 stretchinf(hseq, width - w.getWidth());
      } else {
	 LinkedList<Adapter> items = new LinkedList<Adapter>();
	 boolean first = true; boolean uniform = true; int firstmax = 0;
	 int amount; int maximalAmount;
	 if (w.getWidth() < width) {
	    for (Item item: hseq.getGlueItems()) {
	       Adapter adapter = new Stretcher(item);
	       items.addLast(adapter);
	       if (first) {
		  first = false; firstmax = adapter.getMaximalAmount();
	       } else if (adapter.getMaximalAmount() != firstmax) {
		  uniform = false;
	       }
	    }
	    amount = width - w.getWidth();
	    maximalAmount = w.getStretchability();
	    if (maximalAmount < amount) {
	       maximalAmount = amount;
	    }
	 } else {
	    for (Item item: hseq.getGlueItems()) {
	       Adapter adapter = new Shrinker(item);
	       items.addLast(adapter);
	       if (first) {
		  first = false; firstmax = adapter.getMaximalAmount();
	       } else if (adapter.getMaximalAmount() != firstmax) {
		  uniform = false;
	       }
	    }
	    amount = w.getWidth() - width;
	    maximalAmount = w.getShrinkability();
	 }
	 if (items.size() > 0) {
	    if (uniform) {
	       adaptUniformely(items, items.size(), amount);
	    } else {
	       adaptWeighted(items, items.size(), amount, maximalAmount);
	    }
	 }
      }
   }
} // class HorizontalFitter
