package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * Implementation of the LineBreaker interface that breaks paragraphs
 * into lines using the total-fit algorithm.
 */
public class TotalFitLineBreaker extends BasicLineBreaker {

   public static class Constructor
	 implements LineBreakerFactory.Constructor {
      public LineBreaker create() {
	 return new TotalFitLineBreaker();
      }
   } // class Constructor

   private static int computeBadness(Width width, int parwidth,
	 Item bp, boolean wasflagged) {
      double ratio;
      int beta;

      if (width.getWidth() < parwidth) {
	 if (width.getStretchability() == Item.INFINITY) {
	    ratio = 0;
	 } else if (width.getStretchability() > 0) {
	    ratio = ((double) parwidth - width.getWidth()) /
	       width.getStretchability();
	 } else {
	    // this is bad, but not as bad as shrinking over
	    // the the limits of shrinkability
	    ratio = 100 + 2 * (parwidth - width.getWidth());
	 }
      } else {
	 if (width.getShrinkability() > 0) {
	    ratio = ((double) width.getWidth() - parwidth) /
	       width.getShrinkability();
	    if (ratio > 1) {
	       // do not shrink beyond the given shrinkability
	       ratio = Item.INFINITY;
	    }
	 } else {
	    // this is bad but probably unavoidable
	    ratio = 200 + 2 * (width.getWidth() - parwidth);
	 }
      }
      if (ratio >= Item.INFINITY) {
	 beta = Item.INFINITY;
      } else {
	 beta = (int) (100 * ratio * ratio * ratio + 0.5);
	 if (beta > Item.INFINITY) {
	    beta = Item.INFINITY;
	 }
      }
      if (beta == Item.INFINITY) {
	 return Item.INFINITY;
      }

      int delta = beta + 1;
      int penalty = bp.getPenalty();
      if (penalty > 0) {
	 delta += penalty;
      }
      delta *= 2;
      if (penalty < 0 && penalty > -Item.INFINITY) {
	 delta -= penalty * penalty;
      }
      if (bp.getPenaltyFlag() && wasflagged) {
	 delta += 3000;
      }
      if (delta > Item.INFINITY) {
	 delta = Item.INFINITY;
      }
      assert(delta >= 0 || penalty < 0);
      return delta;
   }

   /**
    * Instances of this private class represent feasible breakpoints.
    */
   private class Candidate {
      public LinkedList<HorizontalSequence> lines;
      public int badness;
      public HorizontalSequence line;
      int parwidth;
      boolean flagged;
      public Candidate parent;
      public boolean weak;
      public boolean survived;

      public Candidate(int parwidth) {
	 lines = new LinkedList<HorizontalSequence>();
	 line = null;
	 badness = 0;
	 this.parwidth = parwidth;
	 flagged = false;
	 parent = null;
	 weak = false;
	 survived = true;
      }

      public Candidate(Candidate old, Item bp) {
	 if (old.weak) {
	    parent = old.parent;
	    weak = true;
	 } else {
	    parent = old;
	    weak = !old.within(bp);
	 }
	 parwidth = old.parwidth;
	 lines = new LinkedList<HorizontalSequence>(old.lines);
	 badness = old.badness;
	 HorizontalSequence newline = old.line.clone();
	 Item hyphen = bp.getHyphen();
	 if (hyphen != null) {
	    newline.add(hyphen);
	 }
	 lines.addLast(newline);
	 badness += computeBadness(newline.getWidth(), parwidth,
	    bp, old.flagged);
	 if (badness > Item.INFINITY) {
	    badness = Item.INFINITY;
	 }
	 flagged = bp.getPenaltyFlag();
	 line = null;
	 survived = true;
      }

      public void finish(Item bp) {
	 if (line != null) {
	    line.add(bp);
	    badness += computeBadness(line.getWidth(), parwidth,
	       bp, flagged);
	    if (badness > Item.INFINITY) {
	       badness = Item.INFINITY;
	    }
	    lines.addLast(line);
	    line = null;
	    flagged = bp.getPenaltyFlag();
	 }
      }

      public void add(Item item) {
	 if (line == null) {
	    line = new SimpleHorizontalSequence();
	 }
	 line.add(item);
      }

      public void add(HorizontalSequence hseq) {
	 if (line == null) {
	    line = new SimpleHorizontalSequence();
	 }
	 line.add(hseq);
      }

      public void add(Item bp, HorizontalSequence word) {
	 if (line == null || bp == null) {
	    add(word);
	 } else {
	    add(bp); add(word);
	 }
      }

      public boolean fittable(Item bp) {
	 if (line == null) {
	    return true;
	 }
	 Width width = line.getWidth();
	 Item hyphen = bp.getHyphen();
	 if (hyphen != null) {
	    width.add(hyphen);
	 }
	 return width.getWidth() - width.getShrinkability() <= parwidth;
      }

      public boolean within(Item bp) {
	 if (line == null) {
	    return false;
	 }
	 Width width = line.getWidth();
	 Item hyphen = bp.getHyphen();
	 if (hyphen != null) {
	    width.add(hyphen);
	 }
	 return width.getWidth() - width.getShrinkability() <= parwidth &&
	        width.getWidth() + width.getStretchability() >= parwidth;
      }

      public String toString() {
	 StringBuffer buffer = new StringBuffer("");
	 buffer.append("candidate consisting of " + lines.size() + " lines");
	 buffer.append(" of badness " + badness);
	 if (line == null) {
	    buffer.append(" and an empty line");
	 } else {
	    buffer.append(" and following text: \"");
	    for (Item item: line) {
	       int code = item.getChar();
	       if (code >= 0) {
		  buffer.append((char) code);
	       }
	    }
	    buffer.append("\"");
	 }
	 return buffer.toString();
      }
   } // class Candidate

   private static String getString(HorizontalSequence hseq) {
      StringBuffer buffer = new StringBuffer("");
      for (Item item: hseq) {
	 int code = item.getChar();
	 if (code >= 0) {
	    buffer.append((char) code);
	 }
      }
      return buffer.toString();
   }

   /**
    * Apply the total-fit line-breaking algorithm to the given
    * horizontal sequence and return a vertical box with the
    * individual lines.
    * @param hseq The horizontal sequence of boxes that
    *             is to be split into lines.
    * @param parwidth The width of the paragraph in units (1/1000 points).
    * @param baselineskip The vertical distance between the baselines of
    *                     two subsequent lines in units (1/1000 points).
    */
   public Item breakParagraph(HorizontalSequence hseq,
	 int parwidth, int baselineskip) {
      LinkedList<Candidate> candidates = new LinkedList<Candidate>();
      candidates.addLast(new Candidate(parwidth));

      Item bp = null; // last breakpoint
      for (HorizontalSequence word: hseq.getWords()) {
	 Item nextbp = word.getFollowingBreakpoint();
	 Candidate favorite = null; int badness = 0;
	 LinkedList<Candidate> survivors = new LinkedList<Candidate>();
	 for (Candidate candidate: candidates) {
	    candidate.add(bp, word);
	    if (candidate.fittable(nextbp) &&
		  (!candidate.weak || !candidate.parent.survived)) {
	       survivors.addLast(candidate);
	       Candidate survivor = new Candidate(candidate, nextbp);
	       int newbadness = survivor.badness;
	       if (favorite == null || newbadness < badness) {
		  favorite = survivor; badness = newbadness;
	       }
	    } else {
	       candidate.survived = false;
	    }
	 }
	 if (favorite != null) {
	    survivors.addLast(favorite);
	 }
	 if (survivors.size() > 0) {
	    candidates = survivors;
	 }
	 bp = nextbp;
      }
      Candidate favorite = null; int badness = 0;
      for (Candidate candidate: candidates) {
	 candidate.finish(bp);
	 if (favorite == null || candidate.badness < favorite.badness) {
	    favorite = candidate; badness = favorite.badness;
	 }
      }
      VerticalBox vbox = new VerticalBox(baselineskip);
      for (HorizontalSequence line: favorite.lines) {
	 add(vbox, line, parwidth);
      }
      return vbox;
   }
} // class TotalFitLineBreaker
