package de.uniulm.mathematik.tries;

import java.lang.*;
import java.util.*;

/**
 * Trie implemenation that supports the interfaces of TrieReader
 * and TrieConstructor.
 */
public class HashedTrie<TrieInfo>
   implements TrieReader<TrieInfo>, TrieConstructor<TrieInfo> {

   /**
    * Node of a hashed trie that represents a sub-trie.
    */
   public class TrieNode<TrieInfo> implements TrieIterator<TrieInfo> {
      protected TrieInfo info;
      protected HashMap<Integer, TrieNode<TrieInfo> > subnodes;

      protected TrieNode() {
	 info = null;
	 subnodes = new HashMap<Integer, TrieNode<TrieInfo> >();
      }

      /**
       * Move downward to the sub-trie associated with code.
       * @param code The next codepoint of the key.
       * @return pointer to node representing the sub-trie,
       *         null is returned if no such sub-trie exists
       */
      public TrieNode<TrieInfo> descend(int code) {
	 return subnodes.get(new Integer(code));
      }

      /**
       * Returns TrieInfo object associated with this node.
       * Null is returned if no object is associated with it.
       */
      public TrieInfo getInfo() {
	 return info;
      }

      /**
       * Return iterator that allows to iterate through all
       * code points that lead to further sub-tries.
       */
      public Iterator<Integer> iterator() {
	 return subnodes.keySet().iterator();
      }
   }

   private TrieNode<TrieInfo> root;
   private int numberOfNodes;
   private int numberOfEntries;

   /**
    * Create a hashed trie.
    */
   public HashedTrie() {
      root = new TrieNode<TrieInfo>();
      numberOfNodes = 0; numberOfEntries = 0;
   }

   /**
    * Return root node of this trie which is always non-null.
    */
   public TrieNode<TrieInfo> getRoot() {
      return root;
   }

   /**
    * Insert the info object into this trie using word as key.
    * @param word Key string which is interpreted as a sequence
    *             of code points.
    * @param info Object to be stored at the given key, must not
    *             be null.
    */
   public void insert(String word, TrieInfo info) {
      /* iterate through all code points of word
	 (which are possibly 32 bit values, not just 16 bit as char) */
      assert info != null;
      TrieNode<TrieInfo> node = root;
      for (int index = 0; index < word.length();
	    index = word.offsetByCodePoints(index, 1)) {
	 int ch = word.codePointAt(index);
	 Integer key = new Integer(ch);
	 TrieNode<TrieInfo> subnode = node.subnodes.get(key);
	 if (subnode == null) {
	    subnode = new TrieNode<TrieInfo>();
	    node.subnodes.put(key, subnode);
	    ++numberOfNodes;
	 }
	 node = subnode;
      }
      node.info = info;
      ++numberOfEntries;
   }

   /**
    * Return number of info objects stored into this trie.
    */
   public int getNumberOfEntries() {
      return numberOfEntries;
   }

   /**
    * Return number of nodes within this trie.
    */
   public int getNumberOfNodes() {
      return numberOfNodes;
   }
}
